
# Standard
from functools import reduce
from typing import Optional, List

# 3rd Party
import numpy as np

# Local
from sampler.boundary_conditions.base import (
    BoundaryConditions,
    BoundaryCondition
)
from sampler.utils.type_aliases import (
    NDArray
)


class UnitSquareBoundaryConditions(BoundaryConditions):


    def __init__(
        self, *,
        top:Optional[BoundaryCondition],
        bot:Optional[BoundaryCondition],
        left:Optional[BoundaryCondition],
        right:Optional[BoundaryCondition],
        x_dim:int,
        y_dim:int,
        consistency_rtol:float = 1.e-5
    ):

        self.field_shape = (x_dim, y_dim)

        # Ravel the 2D indices to 1D global indices
        # NOTE:
        #   multi-index ravel does not seem to work with '-1', and
        #   it is necessary to use N-1 to ravel the last index value.
        self.bot   = self.__init_boundary_condition(bot,   x_idx=np.arange(x_dim),            y_idx=np.zeros((x_dim,)))
        self.top   = self.__init_boundary_condition(top,   x_idx=np.arange(x_dim),            y_idx=np.ones((x_dim,))*(y_dim-1))
        self.left  = self.__init_boundary_condition(left,  x_idx=np.zeros((y_dim,)),          y_idx=np.arange(y_dim))
        self.right = self.__init_boundary_condition(right, x_idx=np.ones((y_dim,))*(x_dim-1), y_idx=np.arange(y_dim))

        # Confirm that there are no inconsistent boundary conditions
        self.problem_corners = self.__get_problem_corners(rtol=consistency_rtol)
        if not self.are_consistent:
            raise RuntimeError(
                "The provided boundary conditions do not agree at the %s corner(s) "
                "of the unit square. Aborting!" % ', '.join(self.problem_corners)
            )

        self.__values, self.__indices = None, None
        if self:
            # Repeated indices correspond to corners, which are already required
            # to be consistent, so we can remove them.
            joint_idx = np.hstack([bc.indices for bc in self if bc is not None])
            joint_value = np.hstack([bc.value for bc in self if bc is not None])
            _, unique_idx = np.unique(joint_idx, return_index=True)
            #
            self.__values = joint_value[unique_idx]
            self.__indices = joint_idx[unique_idx]


    @property
    def are_consistent(self):
        return len(self.problem_corners) == 0


    @property
    def values(self) -> NDArray:
        return self.__values


    @property
    def indices(self) -> NDArray:
        return self.__indices


    def __bool__(self):
        return reduce(
            lambda has_bcs, bc: has_bcs or bc is not None,
            self,
            False
        )


    def __get_problem_corners(
            self,
            rtol:float = 1.e-5
        ) -> List[str]:

        # NOTE:
        #   - bcs are the boundary conditions that overlap
        #   - idx are the indices where boundary conditions overlap
        #   - key describes the location where overlap occurs
        bcs = [(self.left, self.top), (self.right, self.top), (self.left, self.bot), (self.right, self.bot)]
        idx = [(-1, 0), (-1, -1), (0, 0), (0, -1)]
        key = ["top-left", "top-right", "bot-left", "bot-right"]

        problem_corners = []
        for b, i, k in zip(bcs, idx, key):
            # NOTE:
            #   If either boundary condition is None, then 
            #   a conflict is impossible. Otherise, check
            #   that the overlapping values agree to within rtol
            bcs_are_consistent = (
                    b[0] is None
                or  b[1] is None
                or  np.isclose(b[0].value[i[0]], b[1].value[i[1]], rtol=rtol)
            )
            
            if not bcs_are_consistent:
                problem_corners.append(k)
        #
        return problem_corners


    def __init_boundary_condition(
            self,
            boundary_condition:BoundaryCondition,
            x_idx:NDArray,
            y_idx:NDArray
        ):

        if boundary_condition is not None:
            boundary_condition.indices = np.ravel_multi_index(
                (
                    x_idx.astype(int),
                    y_idx.astype(int)
                ),
                self.field_shape
            )

        return boundary_condition


    def __iter__(self):
        return iter([self.top, self.bot, self.left, self.right])




# Standard
from typing import Optional

# 3rd Party
import numpy as np
import scipy as sp

# Local
from sampler.boundary_conditions.base import BoundaryConditions
from sampler.samplers.base import __Sampler__
from sampler.utils.type_aliases import (
    NDArray
)

class GRFSampler(__Sampler__):
    """Gaussian Random Field Sampler"""


    def __init__(
        self,
        average:NDArray,
        cov_mat:NDArray,
        boundary_conditions:Optional[BoundaryConditions],
        seed:Optional[int]=None
    ):

        # If desired, fix the random seed
        if seed is not None:
            np.random.seed(seed)

        self.average = average
        self.cov_mat = cov_mat
        self.boundary_conditions = boundary_conditions
        self.fixed_indices:NDArray = [] if boundary_conditions is None else boundary_conditions.indices
        self.free_indices:NDArray = np.setdiff1d(
            np.arange(average.size),
            self.fixed_indices
        )

        # NOTE:
        #   In general, the indices of block `a` are free,
        #   and the indices of block `b` are fixed at the 
        #   assumed boundary conditions.
        self.mu_a = self.average[self.free_indices]

        # Blocks of the covariance matrix, Gamma = [Gamma_aa, Gamma_ab; Gamma_ab.T, Gamma_bb]
        self.Gamma_aa = self.cov_mat[self.free_indices, :][:, self.free_indices]

        # NOTE: If there are no boundary conditions, then
        # the conditional mean is just the mean, and the
        # conditional covariance is just the covariance.
        self.mu_cond = self.mu_a
        self.Gamma_cond = self.Gamma_aa
        if self.boundary_conditions:

            self.mu_b = self.average[self.fixed_indices]
            self.Gamma_ab = self.cov_mat[self.free_indices, :][:, self.fixed_indices]
            self.Gamma_bb = self.cov_mat[self.fixed_indices, :][:, self.fixed_indices]

            # Compute the conditional covariance:
            #   Gamma_cond = Gamma_aa - Gamma_ab * Gamma_bb^-1 * Gamma_ab.T
            tmp_Gamma = np.linalg.solve(self.Gamma_bb, self.Gamma_ab.T)
            self.Gamma_cond = self.Gamma_aa - self.Gamma_ab @ tmp_Gamma

            # Compute the conditional mean:
            #   mu_cond = mu_a + Gamma_ab * Gamma_bb^-1 * (x_b - mu_b)
            tmp_mu = np.linalg.solve(self.Gamma_bb, (self.boundary_conditions.values - self.mu_b))
            self.mu_cond = self.mu_a + self.Gamma_ab @ tmp_mu

        # Get the Cholesky decomposition of the conditional covariance
        self.L_cond = np.linalg.cholesky(self.Gamma_cond)


    def __call__(self, num_samples:int, std_normal_sample:NDArray = None) -> np.ndarray:
        """Retrieves num_samples samples"""

        if std_normal_sample is None:
            std_normal_sample = np.random.standard_normal((num_samples, self.free_indices.size))
        else:
            num_samples = std_normal_sample.shape[0]

        v = (self.L_cond @ std_normal_sample.T).T

        out_samples = np.empty((num_samples, self.average.size))
        out_samples[:, self.free_indices] = self.mu_cond + v

        # If there are boundary conditions, apply them
        if self.boundary_conditions:
            out_samples[:, self.fixed_indices] = self.boundary_conditions.values

        return out_samples



# 3rd Party
import numpy as np
import scipy as sp
from hypothesis import given, settings

# Local
from sampler.utils.test import (
    is_positive_definite,
    positive_definite_matrix,
    converge_to_same_distribution
)
from sampler.boundary_conditions import (
    UnitSquareBoundaryConditions,
    DirichletBC
)
from sampler.utils.type_aliases import (
    NDArray,
    SparseMatrix
)
from sampler.samplers import(
    GMRFSampler,
    GRFSampler
)


@settings(max_examples=10, deadline=None)
@given(positive_definite_matrix)
def test_grf_gmrf_consistency(precision:SparseMatrix):

    assert is_positive_definite(precision.toarray())

    num_nodes = precision.shape[0]
    average = np.random.rand(num_nodes)*10. - 5.
    covariance = sp.sparse.linalg.inv(precision).toarray()

    one_dim = int(np.sqrt(num_nodes))
    top_values = (np.random.rand(one_dim)*10. - 5.)

    bc_gmrf = UnitSquareBoundaryConditions(
        top = DirichletBC(value=top_values),
        bot = None,
        left = None,
        right = None,
        x_dim=one_dim,
        y_dim=one_dim
    )
    gmrf = GMRFSampler(
        average=average,
        prec_mat=precision,
        boundary_conditions=bc_gmrf,
        use_cuthill_mckee=False
    )


    bc_grf = UnitSquareBoundaryConditions(
        top = DirichletBC(value=top_values),
        bot = None,
        left = None,
        right = None,
        x_dim=one_dim,
        y_dim=one_dim
    )
    grf = GRFSampler(
        average=average,
        cov_mat=covariance,
        boundary_conditions=bc_grf
    )

    # First confirm that the computed means are the same,
    # before checking that the sample means and sample cov
    # agree.
    assert np.allclose(gmrf.mu_b, grf.mu_b)
    assert np.allclose(gmrf.mu_a, grf.mu_a)
    assert np.allclose(gmrf.mu_cond, grf.mu_cond)
    #
    assert converge_to_same_distribution(gmrf, grf)


@settings(max_examples=10, deadline=None)
@given(positive_definite_matrix)
def test_gmrf_permutation_consistency(precision:SparseMatrix):

    assert is_positive_definite(precision.toarray())

    num_nodes = precision.shape[0]
    average = np.random.rand(num_nodes)*10. - 5.

    one_dim = int(np.sqrt(num_nodes))
    left_values = (np.random.rand(one_dim)*10. - 5.)

    bc_gmrf_1 = UnitSquareBoundaryConditions(
        top = None,
        bot = None,
        left = DirichletBC(value=left_values),
        right = None,
        x_dim=one_dim,
        y_dim=one_dim
    )
    gmrf_1 = GMRFSampler(
        average=average,
        prec_mat=precision,
        boundary_conditions=bc_gmrf_1,
        use_cuthill_mckee=False
    )


    bc_gmrf_2 = UnitSquareBoundaryConditions(
        top = None,
        bot = None,
        left = DirichletBC(value=left_values),
        right = None,
        x_dim=one_dim,
        y_dim=one_dim
    )
    gmrf_2 = GMRFSampler(
        average=average,
        prec_mat=precision,
        boundary_conditions=bc_gmrf_2,
        use_cuthill_mckee=True
    )

    # First confirm that the computed means are the same,
    # before checking that the sample means and sample cov
    # agree.
    assert np.allclose(gmrf_1.mu_b, gmrf_2.mu_b)
    assert np.allclose(gmrf_1.mu_a, gmrf_2.mu_a)
    assert np.allclose(gmrf_1.mu_cond, gmrf_2.mu_cond)
    #
    assert converge_to_same_distribution(gmrf_1, gmrf_2)



# Standard
from dataclasses import dataclass
from typing import Optional, Hashable
from abc import ABC

# 3rd Party
from numpy.typing import ArrayLike

# Local
from sampler.utils.type_aliases import BoundaryConditionFunction, NDArray, Indices


@dataclass
class BoundaryCondition:
    """Container for a boundary condition
    
    Properties
    ==========
    func (BoundaryConditionFunction):
        a function that maps the (x,y) coordinates of the boundary
        to the value of gradient(solution).dot(outward_pointing_normal)
    value (ArrayLike):
        an array of boundary condition values; note that value.size must be
        equal to indices[0].size
    indices (ArrayLike):
        the indices of the GRF at which the boundary condition values apply;
        note that indices[0].size must be equal to value.size
    """
    func:Optional[BoundaryConditionFunction] = None
    value:Optional[ArrayLike] = None
    indices:Optional[Indices] = None
    id:Optional[Hashable] = "no_id_provided"


@dataclass
class NeumannBC(BoundaryCondition):
    """Convience class to distinguish types of boundary conditions"""


@dataclass
class DirichletBC(BoundaryCondition):
    """Convience class to distinguish types of boundary conditions"""


class BoundaryConditions(ABC):
    """Abstract container for holding boundary conditions"""

    @property
    def values(self) -> NDArray:
        """Returns the values corresponding to self.indices"""

    @property
    def indices(self) -> NDArray:
        """Returns the indices of the boundary conditions"""

    @property
    def are_consistent(self) -> bool:
        """Returns true if the boundary conditions agree at the intersection of their domains"""



import numpy as np
import scipy as sp

def whittle_matern_precision(num_bases_one_d, length_scale):
    """Generate a precision matrix for a Whittle-Matern covariance function on a uniform 2D grid.

    This precision matrix is valid for a uniform grid with smoothness parameter 2.0.

    Arguments
    =========
    num_bases (int)
        Number of basis functions in the x-direction and y-direction.
    length_scale (float)
        Length scale of the covariance function.

    Returns
    =======
        scipy.sparse.csc_matrix: Precision matrix for the Whittle-Matern covariance function.
    """
    total_num_bases = num_bases_one_d * num_bases_one_d

    eye = sp.sparse.eye(num_bases_one_d)
    L2 = sp.sparse.diags(
        diagonals = (
            -2*np.ones(num_bases_one_d),
            np.ones(num_bases_one_d-1),
            np.ones(num_bases_one_d-1)
        ),
        offsets = (0, -1, 1),
        shape = (num_bases_one_d, num_bases_one_d)
    )

    scaling = 1./(length_scale**2.)
    factor = (
            scaling * sp.sparse.eye(total_num_bases, format="csc")
        -   sp.sparse.kron(eye, L2)
        -   sp.sparse.kron(L2, eye)
    )

    # NOTE:
    #   If the code were structured differently, we could
    #   just pass the factor to the sampler directly--no need to
    #   compute the precision matrix and then do a Cholesky
    #   factorization. However, I wanted to make the sampler
    #   more general, so it accepts a precision matrix.
    return sp.sparse.csc_matrix(factor @ factor.T)


def whittle_matern_covariance(
        num_bases_one_d,
        length_scale,
        variance_scale=1.
    ):
    """Generate a precision matrix for a Whittle-Matern covariance function on a uniform 2D grid.

    The smoothness parameter is fixed at 2.5

    Arguments
    =========
    num_bases (int)
        Number of basis functions in the x-direction and y-direction.
    length_scale (float)
        Length scale of the covariance function.
    scaling_parameter (float)
        Scaling parameter for the covariance function.
    Returns
    =======
        nump.ndarray: Covariance matrix for the Whittle-Matern covariance function.
    """
    # NOTE:
    #   The smoothness parameter is fixed at 2.5, but this function
    #   can be generalized. I just wanted to keep things simple for now.
    #
    disc = np.linspace(0, 1, num_bases_one_d)
    X, Y = np.meshgrid(disc, disc)
    x = np.atleast_2d(X.flatten())
    y = np.atleast_2d(Y.flatten())
    dist = np.sqrt((x - x.T)**2 + (y - y.T)**2)
    scaled_dist = np.sqrt(5) * dist / length_scale

    out_cov = (
            variance_scale
        *   np.multiply(
                (
                        1.
                    +   scaled_dist
                    +   (np.power(scaled_dist, 2.)/3.)
                ),
                np.exp(-scaled_dist)
        )
    )

    return out_cov



import numpy as np
import torch

numpy_to_torch_dtype_dict = {
    np.uint8      : torch.uint8,
    np.int8       : torch.int8,
    np.int16      : torch.int16,
    np.int32      : torch.int32,
    np.int64      : torch.int64,
    np.float16    : torch.float16,
    np.float32    : torch.float32,
    np.float64    : torch.float64,
    np.complex64  : torch.complex64,
    np.complex128 : torch.complex128
}



# Standard
from typing import Tuple

# 3rd Party
import scipy as sp
import numpy as np

# Local
from sampler.utils.type_aliases import (
    SparseMatrix,
    NDArray
)

def chol(A:SparseMatrix, use_cuthill_mckee:bool=True) -> Tuple[SparseMatrix, NDArray]:
    """Cholesky decomposition of a sparse matrix A.

    Computes a sparse Cholesky decomposition PAP^T = LL^T,
    where P is the reverse cuthill mckee permutation.

    Arguments
    =========
    A (SparseMatrix)
        Sparse, symmetric positive definite matrix.

    Returns
    =======
    L (SparseMatrix)
        Lower triangular matrix.
    P (SparseMatrix)
        Permutation matrix.
    """

    # NOTE:
    #   The Reverse Cuthill-McKee permutation is used to reduce the
    #   profile (not bandwidth) of the matrix. This decreases the
    #   density of the resulting Cholesky factor.
    P = sp.sparse.eye(A.shape[0])
    if use_cuthill_mckee:
        p = sp.sparse.csgraph.reverse_cuthill_mckee(A, symmetric_mode=True)
        P = sp.sparse.csc_array(
            (
                np.ones(A.shape[0]),        # data
                (
                    np.arange(A.shape[0]),  # row idx
                    p                       # col idx
                )
            ),
            shape=A.shape
        )

    return sp.sparse.csc_array(np.linalg.cholesky((P@A@P.T).toarray())), P


# Standard
from typing import Union, Tuple, Callable

# 3rd Party
import numpy as np
import scipy as sp
from numpy.typing import ArrayLike
from nutils import topology as nutils_topology
from nutils import function as nutils_function

# Local
# N/A


NutilsFunctionArray = nutils_function.Array
NutilsTopology = nutils_topology.Topology
NDArray = np.ndarray
SparseMatrix = sp.sparse.spmatrix
ArrayPlaceHolder = Union[NDArray, bool, None]
BoundaryConditionFunction = Callable[[ArrayLike], ArrayLike]
Indices = Tuple[ArrayLike, ...]




# Standard
from typing import Optional

# 3rd Party
import numpy as np
import scipy as sp

# Local
from sampler.boundary_conditions.base import BoundaryConditions
from sampler.utils.sparse_chol import chol
from sampler.utils.type_aliases import (
    NDArray,
    SparseMatrix
)
from sampler.samplers.base import __Sampler__


class GMRFSampler(__Sampler__):
    """Gaussian Markov Random Field Sampler"""


    def __init__(
        self, *,
        average:NDArray,
        prec_mat:SparseMatrix,
        boundary_conditions:Optional[BoundaryConditions],
        use_cuthill_mckee:bool=True,
        seed:Optional[int]=None
    ) -> None:

        # If desired, fix the random seed
        if seed is not None:
            np.random.seed(seed)

        self.average = average
        self.prec_mat = prec_mat
        self.boundary_conditions = boundary_conditions
        self.fixed_indices:NDArray = [] if boundary_conditions is None else boundary_conditions.indices
        self.free_indices:NDArray = np.setdiff1d(
            np.arange(average.size),
            self.fixed_indices
        )

        # NOTE:
        #   In general, the indices of block `a` are free,
        #   and the indices of block `b` are fixed at the 
        #   assumed boundary conditions.
        self.mu_a = self.average[self.free_indices]
        
        # NOTE:
        #   The conditonal precision matrix does not need
        #   to be computed, because it's simply the submatrix
        #   extracted from the free indices.
        self.Q_aa = self.prec_mat[self.free_indices, :][:, self.free_indices]
        
        # NOTE:
        #   The cholesky factorization of a sparse matrix is
        #   not necessarily spartse. To increase the sparsity
        #   of the cholesky factor, we permute the rows and
        #   columns using the Reverse Cuthill-McKee ordering,
        #   which is returned as P_aa:
        #     P_aa @ Q_aa @ P_aa.T =  G_aa @ G_aa.T 
        self.G_aa, self.P_aa = chol(self.Q_aa, use_cuthill_mckee=use_cuthill_mckee)
        self.L_aa = self.P_aa.T @ self.G_aa

        # NOTE: If there are no boundary conditions, then
        # the conditional mean is just the mean.
        self.mu_cond = self.mu_a
        if self.boundary_conditions:

            self.mu_b = self.average[self.fixed_indices]
            self.Q_ab = self.prec_mat[self.free_indices, :][:, self.fixed_indices]

            # NOTE:
            #   This is the conditional beta for the canonical GMRF
            #   (x_a - mu_a) | x_b ~ Nc(beta_cond, Q_aa)
            beta_cond = self.Q_ab @ (self.mu_b - self.boundary_conditions.values)
            w = sp.sparse.linalg.spsolve(self.L_aa, beta_cond)
            # NOTE:
            #   We solve for the conditional mean of (x_a - mu_a) | x_b,
            #   and then add mu_a back to get the conditional mean of x_a | x_b.
            self.mu_cond = sp.sparse.linalg.spsolve(self.L_aa.T, w) + self.mu_a


    def __call__(self, num_samples:int, std_normal_sample:NDArray=None) -> NDArray:

        if std_normal_sample is None:
            std_normal_sample = np.random.standard_normal((num_samples, self.free_indices.size))
        else:
            num_samples = std_normal_sample.shape[0]

        v = sp.sparse.linalg.spsolve(self.L_aa.T, std_normal_sample.T).T

        out_samples = np.empty((num_samples, self.average.size))
        out_samples[:, self.free_indices] = self.mu_cond + v

        # If there are boundary conditions, apply them
        if self.boundary_conditions:
            out_samples[:, self.fixed_indices] = self.boundary_conditions.values

        return out_samples



import numpy as np
import scipy as sp
from tabulate import tabulate
from hypothesis.strategies import integers
from sampler.samplers.base import __Sampler__
from sampler.utils.type_aliases import (
    NDArray,
    SparseMatrix
)


def converges_to_distribution(
        sampler:__Sampler__,
        average:NDArray,
        covariance:NDArray,
        max_iter=8,
        rtol:float=5e-2,
        atol:float=1e-2
) -> None:
    """Check if sampler converges to specified distribution to within rtol.
    
    Arguments
    ---------
    sampler (__Sampler__)
        A sampler object
    average (NDArray)
        Average of the distribution
    covariance (NDArray)
        Covariance of the distribution
    max_iter (int)
        Maximum number of iterations; each iteration increases
        the sample size by a factor of 100
    rtol (float)
        Relative tolerance
    """
    num_samples = np.power(10, np.arange(1, max_iter)).astype(int)
    converges = False
    avg_residual = []
    cov_residual = []
    for n in num_samples:
        sample = sampler(n)
        sample_covariance = np.cov(sample.T)
        sample_average = np.average(sample, axis=0)

        if (
                np.allclose(sample_average, average, rtol=rtol, atol=atol)
            and np.allclose(sample_covariance, covariance, rtol=rtol, atol=atol)
        ):
            converges = True
            break

        avg_residual.append(
            (np.abs(sample_average - average) - rtol*np.abs(average)).max()
        )
        cov_residual.append(
            (np.abs(sample_covariance - covariance) - rtol*np.abs(covariance)).max()
        )

    if not converges:
        failure_mode = "`cov`" if np.min(cov_residual) > np.min(avg_residual) else "`mu`"
        print(f"The max {failure_mode} relative residual exceeds {atol} after {n} samples")
        print(
            tabulate(
                {
                    'Sample Size': num_samples,
                    'Mu Max Rel. Residual': avg_residual,
                    'Cov. Max Rel. Residual': cov_residual
                },
                headers='keys',
                tablefmt='pretty'
            )
        )

    return converges


def converge_to_same_distribution(
        sampler_1:__Sampler__,
        sampler_2:__Sampler__,
        max_iter=8,
        rtol:float=5e-2,
        atol:float=1e-2
) -> None:
    """Check if samplers converges to same distribution to within rtol.
    
    Arguments
    ---------
    sampler_1 (__Sampler__)
        A sampler object
    sampler_2 (__Sampler__)
        A sampler object
    max_iter (int)
        Maximum number of iterations; each iteration increases
        the sample size by a factor of 10
    rtol (float)
        Relative tolerance
    """
    num_samples = np.power(10, np.arange(1, max_iter)).astype(int)
    converges = False
    avg_residual = []
    cov_residual = []
    for n in num_samples:

        sample_1 = sampler_1(n)
        sample_2 = sampler_2(n)

        sample_1_covariance = np.cov(sample_1.T)
        sample_1_average = np.average(sample_1, axis=0)

        sample_2_covariance = np.cov(sample_2.T)
        sample_2_average = np.average(sample_2, axis=0)

        if (
                np.allclose(sample_1_average, sample_2_average, rtol=rtol, atol=atol)
            and np.allclose(sample_1_covariance, sample_2_covariance, rtol=rtol, atol=atol)
        ):
            converges = True
            break

        avg_residual.append(
            (np.abs(sample_1_average - sample_2_average) - rtol*np.abs(sample_1_average)).max()
        )
        cov_residual.append(
            (np.abs(sample_1_covariance - sample_2_covariance) - rtol*np.abs(sample_1_covariance)).max()
        )

    if not converges:
        failure_mode = "`cov`" if np.min(cov_residual) > np.min(avg_residual) else "`mu`"
        print(f"The max {failure_mode} relative residual exceeds {atol} after {n} samples")
        print(
            tabulate(
                {
                    'Sample Size': num_samples,
                    'Mu Max Rel. Residual': avg_residual,
                    'Cov. Max Rel. Residual': cov_residual
                },
                headers='keys',
                tablefmt='pretty'
            )
        )

    return converges


def is_positive_definite(matrix:NDArray) -> bool:
    """Check if a matrix is positive definite.
    
    Arguments
    ---------
    matrix (NDArray)
        A square matrix

    Returns
    -------
    is_psd (bool)
        True if the matrix is positive definite
    """
    try:
        np.linalg.cholesky(matrix)
        return True
    except np.linalg.LinAlgError:
        return False


# Define a function for filtering out non-positive definite matrices
def get_positive_definite_matrix(size:int) -> SparseMatrix:
    """Check if a matrix is positive definite.
    
    Arguments
    ---------
    size (int)
        Size of each dim of the matrix

    Returns
    -------
    proposed_psd (SparseMatrix)
        A sparse positive definite matrix
    """
    proposed_psd:NDArray
    while True:
        tmp = sp.sparse.random(size, size, density=0.3, format='csr')
        proposed_psd = tmp @ tmp.T
        try:
            np.linalg.cholesky(proposed_psd.toarray())
            break
        except np.linalg.LinAlgError:
            continue

    return sp.sparse.csc_array(proposed_psd)


# Define a strategy for sampling positive definite matrices
positive_definite_matrix = integers(min_value=3, max_value=5) \
    .map(lambda n: get_positive_definite_matrix(n**2)) \
    .filter(lambda A: is_positive_definite(A.toarray()))





import numpy as np
import torch_bspline as tb
from sampler.samplers.unit_square import UnitSquareSampler
from sampler.boundary_conditions.base import NeumannBC, DirichletBC
from sampler.utils.kernels import whittle_matern_precision, whittle_matern_covariance

#
# First, sample

# NOTE:
#   - The average is expected to be a 2D array of shape (num_bases_x, num_bases_y)
#   - You can provide either a precision matrix (`prec_mat=...`)
#     or a covariance matrix (`cov_mat=...`), but not both
#   - The `whittle_matern_covariance` function assumes a smoothness parameter of 2.5
#     (this isn't arbitrary, but a special case for which the kernel is simple)
num_bases_x = 15
num_bases_y = 15
average = np.zeros((num_bases_x, num_bases_y))
cov = whittle_matern_covariance(num_bases_x, 0.3)

# NOTE:
#   - The precision matrix is expected to be a 2D array of shape
#     (num_bases_x*num_bases_y, num_bases_x*num_bases_y)
#   - The Whittle-Matern precision matrix is only valid for uniform grids.
#   - The `whittle_matern_precision` function assumes a smoothness parameter of 2.0
#     (this isn't arbitrary, but a special case for which the precision can be
#     computed in closed form and the resulting precision matrix is very sparse)
# prec = whittle_matern_precision(num_bases_x, 1.)

# Example boundary condition functions
# NOTE:
#   Since the boundary conditions are applied to the
#   edges of the unit-square, they only take 1D args
#   (i.e. x or y) corresponding to which ever variable
#   is free on that boundary.
def sin_bc(x):
    return np.sin(2*np.pi*x)

def constant(in_constant=0.):
    def f(x):
        return in_constant
    return f

def linear(in_slope=1., in_intercept=0.):
    def f(x):
        return in_slope*x + in_intercept
    return f

def zero(x):
    return 0.

# NOTE:
#   Neumann conditions are currently unsupported,
#   but they should be up and running soon. I'll
#   add in linear constraints at the same time.
not_supported = NeumannBC(func=sin_bc, id="neumann")

# NOTE:
#   If you want you reuse the same boundary condition,
#   you'll either need multiple instances; otherwise,
#   the computed indices will be overwritten each time
#   the referenced BC is evaluated.
tmp_sampler = UnitSquareSampler(
    average=average,
    cov_mat=cov,
    #prec_mat=prec,
    poly_order=3,
    bc_top=DirichletBC(func=constant(-1.)),
    bc_bot=DirichletBC(func=constant(1.)),
    bc_left=None,
    bc_right=None,
)

weights = tmp_sampler.sample(1)
tmp_sampler.visualize_sample(sample=weights)




# Standard
import contextlib
import io
from functools import reduce
from typing import Tuple, Union

# 3rd Party
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch_bspline as tb
from nutils import mesh as nutils_mesh

# Local
from sampler.boundary_conditions.base import (
    BoundaryCondition,
    NeumannBC,
    DirichletBC
)
from sampler.boundary_conditions.unit_square import UnitSquareBoundaryConditions
from sampler.samplers.base import (
    __Sampler__,
    Sampler,
)
from sampler.samplers.gmrf import GMRFSampler
from sampler.samplers.grf import GRFSampler
from sampler.utils.dtype_map import numpy_to_torch_dtype_dict
from sampler.utils.type_aliases import (
    NDArray,
    SparseMatrix,
    NutilsFunctionArray,
    NutilsTopology
)


class UnitSquareSampler(Sampler):


    def __init__(
        self, *,
        average:NDArray,
        poly_order:int,
        cov_mat:NDArray = None,
        prec_mat:SparseMatrix = None,
        bc_top:BoundaryCondition = None,
        bc_bot:BoundaryCondition = None,
        bc_left:BoundaryCondition = None,
        bc_right:BoundaryCondition = None,
        dtype:np.dtype = np.float64,
        rtol:float = 1.e-5
    ):
        """Initialize UnitSquareSampler
        
        Required Keyword Arguments
        ==========================
        average (NDArray):
            An array of the average coefficient values at each knot. The dimensions
            of the knot vectors are determined by the size of this argument.
        poly_order (int):
            The polynomial order of the b-spline basis.
        cov_mat (NDArray):
            The covariance matrix of the Gaussian Random Field. The function only
            accepts cov_mat XOR prec_mat. If the field has a known, sparse precision
            matrix, then sampling may be faster using the precision matrix.
        prec_mat (SparseMatrix):
            The precision matrix of the Gaussian Random Field. The function only
            accepts cov_mat XOR prec_mat. If the field does not have a known,
            sparse precision matrix, then it probably makes more sense to pass
            the covariance matrix.
        bc_top (BoundaryCondition):
            The boundary condition to be enforced on the top of the unit square (y=1).
            If None, the value at the top boundary will be random.
        bc_bot (BoundaryCondition):
            The boundary condition to be enforced on the bottom of the unit square (y=0).
            If None, the value at the bottom boundary will be random.
        bc_left (BoundaryCondition):
            The boundary condition to be enforced on the left of the unit square (x=0).
            If None, the value at the left boundary will be random.
        bc_right (BoundaryCondition):
            The boundary condition to be enforced on the right of the unit square (x=1).
            If None, the value at the right boundary will be random.
        rtol (float):
            The relative tolerance with which boundary conditions are forced to be consistent
            at the corners.
        """

        self.poly_order = poly_order

        # NOTE: 
        #   Only storing one matrix: self.mat is either covariance or precision. If
        #   the precision matrix is stored, self.is_gmrf == True 
        self.x_dim, self.y_dim = self.__init_dims(average_coeffs=average)
        self.average = average.flatten()
        self.mat, self.is_gmrf = self.__init_mat(cov_mat=cov_mat, prec_mat=prec_mat)

        self.x_basis, self.y_basis, self.xy_basis = self.__init_basis(
            x_dim=self.x_dim,
            y_dim=self.y_dim,
            poly_order=poly_order,
            dtype=dtype
        )

        self.boundary_conditions = self.__init_boundary_conditions(
            bc_top=bc_top,
            bc_bot=bc_bot,
            bc_left=bc_left,
            bc_right=bc_right,
            poly_order=poly_order,
            rtol=rtol
        )

        self.fixed_indices = self.boundary_conditions.indices
        self.free_indices = np.setdiff1d(np.arange(self.x_dim*self.y_dim), self.fixed_indices)

        self.sample = self.__init_sampler(
            is_gmrf=self.is_gmrf,
            average=self.average,
            mat=self.mat,
            boundary_conditions=self.boundary_conditions
        )


    def __init_boundary_conditions(
        self, *,
        bc_top:BoundaryCondition,
        bc_bot:BoundaryCondition,
        bc_left:BoundaryCondition,
        bc_right:BoundaryCondition,
        poly_order:int,
        rtol:float
    ):
        """Sets the boundary conditions.

        Currently, Neumann boundary conditions are not supported,
        and they raise a RuntimeError.
        """

        # Iterates over boundary conditions, checking for Neumann
        has_neumann_bcs = reduce(
            lambda has_neumann, next_bc: has_neumann or (next_bc is not None and isinstance(next_bc, NeumannBC)),
            [bc_top, bc_bot, bc_left, bc_right],
            False
        )

        if has_neumann_bcs:
            raise RuntimeError("Neumann conditions are not currently supported. Aborting!")
        
        # Create temporary topology, geometry, and basis for projecting the
        # boundary conditions. This is only done for convenience.
        x_topo, x_geom = nutils_mesh.rectilinear([np.linspace(0, 1, self.x_dim-poly_order+1)])
        x_basis = x_topo.basis('spline', degree=poly_order)
        #
        y_topo, y_geom = nutils_mesh.rectilinear([np.linspace(0, 1, self.y_dim-poly_order+1)])
        y_basis = y_topo.basis('spline', degree=poly_order)

        # Capturing the projection solver output because I don't see
        # an argument for "silent" or "verbose", and I find it annoying.
        with contextlib.redirect_stdout(io.StringIO()) as _:
                bc_top   = self.project_onto_boundary(boundary_condition=bc_top,   topo=x_topo, basis=x_basis, geometry=x_geom, poly_order=poly_order)
                bc_bot   = self.project_onto_boundary(boundary_condition=bc_bot,   topo=x_topo, basis=x_basis, geometry=x_geom, poly_order=poly_order)
                bc_left  = self.project_onto_boundary(boundary_condition=bc_left,  topo=y_topo, basis=y_basis, geometry=y_geom, poly_order=poly_order)
                bc_right = self.project_onto_boundary(boundary_condition=bc_right, topo=y_topo, basis=y_basis, geometry=y_geom, poly_order=poly_order)

        out_bcs = UnitSquareBoundaryConditions(
            top   = bc_top,
            bot   = bc_bot,
            left  = bc_left,
            right = bc_right,
            x_dim = len(x_basis),
            y_dim = len(y_basis),
            consistency_rtol=rtol
        )

        return out_bcs


    def __init_basis(
        self, *,
        x_dim:int,
        y_dim:int,
        poly_order:int,
        dtype:np.dtype
    ) -> tb.TensorBasis:

        x_basis = tb.BSpline.uniform(
            lims=(0,1),
            n_segments=x_dim - poly_order,
            degree=poly_order,
            dtype = numpy_to_torch_dtype_dict[dtype]
        )
        y_basis = tb.BSpline.uniform(
            lims=(0,1),
            n_segments=y_dim - poly_order,
            degree=poly_order,
            dtype = numpy_to_torch_dtype_dict[dtype]
        )
        xy_basis = tb.TensorBasis(x_basis, y_basis)

        return x_basis, y_basis, xy_basis


    def __init_dims(self, average_coeffs:NDArray) -> Tuple[int,int]:
        if len(average_coeffs.shape) != 2:
            raise RuntimeError("The array of average coefficients must be 2D. Aborting!")

        return average_coeffs.shape


    def __init_mat(self, cov_mat:NDArray, prec_mat:SparseMatrix) -> Tuple[Union[NDArray, SparseMatrix], bool]:
        is_gmrf = cov_mat is None
        if is_gmrf == (prec_mat is None):
            raise RuntimeError("Must provide exactly one of prec_mat XOR cov_mat. Aborting!")
        
        out_mat = prec_mat if is_gmrf else cov_mat
        if (
                self.x_dim*self.y_dim != out_mat.shape[0]
            or  out_mat.shape[0] != out_mat.shape[1]
        ):
            raise RuntimeError("Precision/covariance matrix must be NxN, where N = average.size. Aborting!")

        return (
            out_mat,
            is_gmrf
        )


    def __init_sampler(
        self, *,
        is_gmrf:bool,
        average:NDArray,
        mat:Union[NDArray, SparseMatrix],
        boundary_conditions:UnitSquareBoundaryConditions
    ) -> __Sampler__:

        out_sampler:__Sampler__ = None
        if is_gmrf:
            out_sampler = GMRFSampler(
                average=average,
                prec_mat=mat,
                boundary_conditions=boundary_conditions
            )
        else:
            out_sampler = GRFSampler(
                average=average,
                cov_mat=mat,
                boundary_conditions=boundary_conditions
            )

        return out_sampler


    @classmethod
    def project_onto_boundary(
        cls, *,
        boundary_condition:BoundaryCondition,
        topo:NutilsTopology,
        basis:NutilsFunctionArray,
        geometry:NutilsFunctionArray,
        poly_order:int
    ) -> BoundaryCondition:
        
        if boundary_condition is None:
            return None

        # NOTE:
        #   If the boundaries are Dirichlet, then we can only ensure
        #   consistency if the values at the corners are 'exact'
        use_exact_boundaries = isinstance(boundary_condition, DirichletBC)

        if (func:= boundary_condition.func) is not None:
            boundary_condition.value = topo.project(
                func(geometry[0]),
                onto=basis,
                geometry=geometry,
                ptype="lsqr",
                degree=poly_order,
                exact_boundaries=use_exact_boundaries
            )

        if boundary_condition.value is None:
            raise RuntimeError("BoundaryCondition[%s] has neither function nor value. Aborting!" % str(boundary_condition.id))
        elif boundary_condition.value.size != basis.size:
            raise RuntimeError("BoundaryCondition[%s].value.size does not match basis.size. Aborting!" % str(boundary_condition.id))

        return boundary_condition


    def visualize_sample(
        self, *,
        cmap:str='bwr',
        fontsize:int=40,
        num_x:int=100,
        num_y:int=101,
        num_contour_levels:int=15,
        sample:NDArray=None,
        title='Func Value',
        ) -> None:

        weights = (sample if sample is not None else self.sample(1)).flatten()

        xy_grid = tb.TensorGrid(
            xs = torch.linspace(0,1,num_x),
            ys = torch.linspace(0,1,num_y),
            x_varies_first=True
        )
        X, Y  = np.meshgrid(xy_grid.xs.numpy(), xy_grid.ys.numpy())

        f = tb.BSplineFunctions(self.xy_basis, torch.tensor(weights))
        Z = f(xy_grid).reshape(X.shape)

        fig = plt.figure()
        ax = fig.add_subplot(111)

        cs = ax.contourf(X, Y, Z, levels=num_contour_levels, cmap=cmap)

        ax.set_title(title, fontsize=fontsize)
        fig.subplots_adjust(right=0.8)
        cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
        fig.colorbar(cs, cax=cbar_ax)

        plt.show()

        return xy_grid




# 3rd Party
import numpy as np
import scipy as sp
from hypothesis import given, settings
from hypothesis.strategies import integers

# Local
from sampler.utils.test import (
    is_positive_definite,
    positive_definite_matrix,
    converges_to_distribution
)
from sampler.utils.type_aliases import (
    NDArray,
    SparseMatrix
)
from sampler.samplers import(
    GMRFSampler,
    GRFSampler
)


@given(positive_definite_matrix)
def test_positive_definiteness(psd:SparseMatrix):
    assert is_positive_definite(psd.toarray())


@settings(max_examples=10, deadline=None)
@given(positive_definite_matrix)
def test_grf_validation_no_bcs(precision:SparseMatrix):

    assert is_positive_definite(precision.toarray())

    num_nodes = precision.shape[0]
    average = np.random.rand(num_nodes)*10. - 5.
    covariance = sp.sparse.linalg.inv(precision).toarray()

    grf = GRFSampler(
        average=average,
        cov_mat=covariance,
        boundary_conditions=None
    )

    assert converges_to_distribution(
        sampler=grf,
        average=average,
        covariance=covariance
    )


@settings(max_examples=10, deadline=None)
@given(positive_definite_matrix)
def test_gmrf_validation_no_permutation_no_bcs(precision:SparseMatrix):

    assert is_positive_definite(precision.toarray())

    num_nodes = precision.shape[0]
    average = np.random.rand(num_nodes)*10. - 5.
    covariance = sp.sparse.linalg.inv(precision).toarray()

    gmrf = GMRFSampler(
        average=average,
        prec_mat=precision,
        boundary_conditions=None,
        use_cuthill_mckee=False
    )

    assert converges_to_distribution(
        sampler=gmrf,
        average=average,
        covariance=covariance
    )


@settings(max_examples=10, deadline=None)
@given(positive_definite_matrix)
def test_gmrf_validation_cuthill_mckee_no_bcs(precision:SparseMatrix):

    assert is_positive_definite(precision.toarray())

    num_nodes = precision.shape[0]
    average = np.random.rand(num_nodes)*10. - 5.
    covariance = sp.sparse.linalg.inv(precision).toarray()

    gmrf = GMRFSampler(
        average=average,
        prec_mat=precision,
        boundary_conditions=None,
        use_cuthill_mckee=True
    )

    assert converges_to_distribution(
        sampler=gmrf,
        average=average,
        covariance=covariance
    )




# Standard
from typing import Tuple

# 3rd Party
import numpy as np
import scipy as sp
import pytest
from hypothesis import given
from hypothesis.strategies import floats, integers, tuples


# Local
from sampler.boundary_conditions import (
    UnitSquareBoundaryConditions,
    DirichletBC
)


@given(
    integers(min_value=4, max_value=25),
    tuples(
        floats(min_value=-10., max_value=10.),
        floats(min_value=-10., max_value=10.),
        floats(min_value=-10., max_value=10.),
        floats(min_value=-10., max_value=10.)
    )
)
def test_bc_consistency(size:int, corner_values:Tuple[float, float, float, float]):
    top_left, top_right, bot_left, bot_right = corner_values

    top_values   = (np.random.rand(size)*10. - 5.)
    bot_values   = (np.random.rand(size)*10. - 5.)
    left_values  = (np.random.rand(size)*10. - 5.)
    right_values = (np.random.rand(size)*10. - 5.)

    top_values[0]   = top_left
    left_values[-1] = top_left

    top_values[-1]   = top_right
    right_values[-1] = top_right

    bot_values[0]   = bot_left
    left_values[0]  = bot_left

    bot_values[-1]  = bot_right
    right_values[0] = bot_right

    boundary_conditions = UnitSquareBoundaryConditions(
        top = DirichletBC(value=top_values),
        bot = DirichletBC(value=bot_values),
        left = DirichletBC(value=left_values),
        right = DirichletBC(value=right_values),
        x_dim=size,
        y_dim=size
    )

    assert boundary_conditions.values.size == (4*(size-1))
    assert boundary_conditions.are_consistent


@given(integers(min_value=4, max_value=25), floats(min_value=-10., max_value=10.))
def test_bc_inconsistency(size:int, corner_value:float):

    top_values   = (np.random.rand(size)*10. - 5.)
    left_values  = (np.random.rand(size)*10. - 5.)

    top_values[0]   = corner_value
    left_values[-1] = corner_value + 1.

    with pytest.raises(RuntimeError):
        _ = UnitSquareBoundaryConditions(
            top = DirichletBC(value=top_values),
            bot = None,
            left = DirichletBC(value=left_values),
            right = None,
            x_dim=size,
            y_dim=size
        )




# Standard
import abc

# 3rd Party
import numpy as np

# Local
from sampler.utils.type_aliases import NDArray


class __Sampler__(abc.ABC):
    """Abstract base class for hidden samplers called in Sampler"""
    
    def __call__(self, num_samples:int) -> NDArray:
        """Retrieves num_samples samples"""


class Sampler(abc.ABC):
    """Abstract base class for samplers.
    
    Generally, will contain a __Sampler__ object that is called
    in Sampler.sample(), but doesn't have to.
    """

    def sample(self, num_samples:int) -> NDArray:
        """Retrieves num_samples samples"""
        pass

    def visualize_sample(self, sample:NDArray) -> None:
        """Visualize a single sample"""
        pass
